mod background;
mod state;
use state::State;
mod stuff;
use stuff::Stuff;
mod constants;
use constants::*;
mod surface;
use surface::Surface;
mod util;
use util::*;

use std::fs;
use std::num::NonZeroU32;

use crate::gsa::Gsa;
use crate::gsa_render_to_screen::render_to_screen;
use crate::gsa_render_to_screen::render_to_window;
use crate::maps::Maps;
use crate::sprite::Sprite;
use crate::tileset::*;
use crate::*;
use clap::crate_version;
use glam::IVec2;
use winit::dpi::PhysicalSize;
use winit::event::ElementState;
use winit::{
    dpi::LogicalSize,
    event::{Event, VirtualKeyCode, WindowEvent},
    event_loop::{ControlFlow, EventLoop},
    window::WindowBuilder,
};


pub(crate) fn run_mapedit() {
    println!("running map edit");

    let tileset_path = "examples/basic/gfx.gif";
    let (tileset, palette) = load_tileset(&fs::read(tileset_path).unwrap());

    let maps_path = "examples/basic/maps.dat";
    let maps = if Path::new(maps_path).exists() {
        postcard::from_bytes(&fs::read(maps_path).unwrap()).unwrap()
    } else {
        Maps::default()
    };

    let mut gsa1 = Gsa {
        sprite: [Sprite::default(); MAX_SPRITES],
        palette,
        bg: Default::default(),
        font: FONT_BOLD,
        pressed: 0,
        released: 0,
        down: 0,
        maps,
        str_bg: 3,
    };

    gsa1.reset_bgs();
    gsa1.reset_sprites();

    if gsa1.map_exists(0) {
        gsa1.load_map(0);
    }

    for i in 0..3 {
        gsa1.bg[i].scroll = IVec2 {
            x: -16 - 32,
            y: -32,
        };
    }

    let mut gsa2 = Gsa {
        sprite: [Sprite::default(); MAX_SPRITES],
        palette,
        bg: Default::default(),
        font: FONT_BOLD,
        pressed: 0,
        released: 0,
        down: 0,
        maps: Maps::default(),
        str_bg: 3,
    };
    gsa2.reset_bgs();
    gsa2.reset_sprites();
    for i in 0..2 {
        gsa2.bg[i].scroll = IVec2 {
            x: -16 - 32,
            y: -32,
        };
    }

    for y in 0..TILESET_SIZE {
        for x in 0..TILESET_SIZE {
            gsa2.bg[0].tiles[x][y] = (x + (y << 8)) as u16;
        }
    }
    //draw_input_window(&mut gsa2, IVec2 { x: 5, y: 3 }, "Create Map");

    let mut stuff = Stuff {
        gsa1,
        gsa2,
        mouse_pos: IVec2::ZERO,
        tile_pos: IVec2::ZERO,
        tile_pos2: IVec2::ZERO,
        left_down: false,
        middle_down: false,
        right_down: false,
        selected_tile: IVec2::ZERO,
        current_layer: 0,
        all_layers: true,
        state: State::Edit,
        input_buf: 0,
        current_map: 0,
        window_tile_size: IVec2::new(1280 / 32, 640 / 32),
    };

    let event_loop = EventLoop::new();
    let size = LogicalSize::new(1280, 640);
    let window = WindowBuilder::new()
        .with_title(format!(
            "Game Skunk Advance Map Editor v{}",
            crate_version!()
        ))
        .with_inner_size(size)
        .with_resize_increments(LogicalSize::new(32, 32))
        .build(&event_loop)
        .unwrap();

    let context = unsafe { softbuffer::Context::new(&window) }.unwrap();
    let mut surface = unsafe { softbuffer::Surface::new(&context, &window) }.unwrap();
    window.request_redraw();

    stuff.gsa2.bg[1].tiles[0][0] = TILE_MARKER;
    stuff.set_state(State::Edit);

    event_loop.run(move |event, _, control_flow| {
        //let stuff = &mut stuff;

        match event {
            Event::WindowEvent { event, .. } => match event {
                WindowEvent::Resized(size) => {
                    if size.width % 32 != 0 || size.height % 32 != 0 {
                        let tw = (size.width as f32 / 32.0).round();
                        let th = (size.height as f32 / 32.0).round();
                        let w = (tw * 32.0) as u32;
                        let h = (th * 32.0) as u32;
                        stuff.window_tile_size.x = tw as i32;
                        stuff.window_tile_size.y = th as i32;
                        stuff.draw_toolbar();
                        let pos = window.outer_position().unwrap();
                        window.set_inner_size(PhysicalSize::new(w, h));
                        window.set_outer_position(pos);
                    }
                }
                WindowEvent::KeyboardInput { input, .. } => {
                    let vk = input.virtual_keycode.unwrap_or(VirtualKeyCode::F24);
                    match vk {
                        VirtualKeyCode::Escape => {
                            //*control_flow = ControlFlow::Exit;
                        }
                        VirtualKeyCode::LShift => {
                            if input.state == ElementState::Pressed {
                                stuff.set_state(State::SelectTile);
                            } else {
                                stuff.set_state(State::Edit);
                            }
                        }
                        VirtualKeyCode::Key0
                        | VirtualKeyCode::Key1
                        | VirtualKeyCode::Key2
                        | VirtualKeyCode::Key3
                        | VirtualKeyCode::Key4
                        | VirtualKeyCode::Key5
                        | VirtualKeyCode::Key6
                        | VirtualKeyCode::Key7
                        | VirtualKeyCode::Key8
                        | VirtualKeyCode::Key9
                        | VirtualKeyCode::A
                        | VirtualKeyCode::B
                        | VirtualKeyCode::C
                        | VirtualKeyCode::D
                        | VirtualKeyCode::E
                        | VirtualKeyCode::F => {
                            if input.state == ElementState::Pressed {
                                stuff.input_buf = ((stuff.input_buf & 0xFFF) << 4) | key_to_num(vk);
                                stuff.update_input_window(IVec2 { x: 15, y: 9 }, stuff.input_buf);
                            }
                        }
                        _ => {}
                    }
                }
                WindowEvent::MouseInput {
                    state: but_state,
                    button,
                    ..
                } => match (but_state, button) {
                    (winit::event::ElementState::Pressed, winit::event::MouseButton::Left) => {
                        if stuff.mouse_pos.x < 16 {
                            match (stuff.mouse_pos.y / 16) as usize {
                                BUT_SAVE => {
                                    println!("saving");
                                    stuff.gsa1.store_map(stuff.current_map);
                                    fs::write(
                                        maps_path,
                                        postcard::to_allocvec(&stuff.gsa1.maps).unwrap(),
                                    )
                                    .unwrap();
                                }
                                BUT_CREATE => {
                                    stuff.input_buf = stuff.current_map;
                                    stuff.set_state(State::NewMapDialog);
                                    stuff.update_input_window(
                                        IVec2 { x: 15, y: 9 },
                                        stuff.input_buf,
                                    );
                                }
                                BUT_LAYER1 => {
                                    stuff.current_layer = 0;
                                    stuff.set_state(stuff.state);
                                }
                                BUT_LAYER2 => {
                                    stuff.current_layer = 1;
                                    stuff.set_state(stuff.state);
                                }
                                BUT_LAYER3 => {
                                    stuff.current_layer = 2;
                                    stuff.set_state(stuff.state);
                                }
                                BUT_LAYERS => {
                                    stuff.all_layers = !stuff.all_layers;
                                    stuff.set_state(stuff.state);
                                }
				BUT_SHIFT_LEFT => {
				    stuff.gsa1.bg[0].shift_left();
				    stuff.gsa1.bg[1].shift_left();
				    stuff.gsa1.bg[2].shift_left();
				}
				BUT_SHIFT_RIGHT => {
				    stuff.gsa1.bg[0].shift_right();
				    stuff.gsa1.bg[1].shift_right();
				    stuff.gsa1.bg[2].shift_right();
				}
				BUT_SHIFT_UP => {
				    stuff.gsa1.bg[0].shift_up();
				    stuff.gsa1.bg[1].shift_up();
				    stuff.gsa1.bg[2].shift_up();				    
				}
				BUT_SHIFT_DOWN => {
				    stuff.gsa1.bg[0].shift_down();
				    stuff.gsa1.bg[1].shift_down();
				    stuff.gsa1.bg[2].shift_down();				    
				}
                                BUT_EXIT => {
                                    println!("exit");
                                    *control_flow = ControlFlow::Exit;
                                }
                                _ => {}
                            }
                        } else {
                            match stuff.state {
                                State::NewMapDialog => {
                                    //todo: not hardcode?
                                    if stuff.mouse_pos.y / 16 == 12 {
                                        if stuff.mouse_pos.x / 16 == 16 {
                                            println!("yes");
                                            stuff.gsa1.store_map(stuff.current_map);
                                            stuff.current_map = stuff.input_buf;
                                            if stuff.gsa1.map_exists(stuff.current_map) {
                                                stuff.gsa1.load_map(stuff.current_map);
                                            } else {
                                                //todo: not reset scrolling
                                                stuff.gsa1.reset_bgs();
                                            }
                                            stuff.set_state(State::Edit);
                                        } else if stuff.mouse_pos.x / 16 == 20 {
                                            println!("no");
                                            stuff.set_state(State::Edit);
                                        }
                                    }
                                }
                                _ => {
                                    stuff.left_down = true;
                                }
                            }
                        }
                    }
                    (winit::event::ElementState::Released, winit::event::MouseButton::Left) => {
                        stuff.left_down = false;
                    }
                    (winit::event::ElementState::Pressed, winit::event::MouseButton::Middle) => {
                        stuff.middle_down = true;
                    }
                    (winit::event::ElementState::Released, winit::event::MouseButton::Middle) => {
                        stuff.middle_down = false;
                    }
                    (winit::event::ElementState::Pressed, winit::event::MouseButton::Right) => {
                        stuff.right_down = true;
                    }
                    (winit::event::ElementState::Released, winit::event::MouseButton::Right) => {
                        stuff.right_down = false;
                    }
                    _ => {}
                },
                WindowEvent::CursorMoved { position, .. } => {
                    let new_pos = IVec2 {
                        x: position.x as i32 / 2,
                        y: position.y as i32 / 2,
                    };
                    let delta = new_pos - stuff.mouse_pos;
                    if stuff.middle_down {
                        match stuff.state {
                            State::Edit => {
                                // normal mode
                                stuff.gsa1.bg[0].scroll -= delta;
                                stuff.gsa1.bg[1].scroll -= delta;
                                stuff.gsa1.bg[2].scroll -= delta;
                            }
                            State::SelectTile => {
                                // tile select mode
                                stuff.gsa2.bg[0].scroll -= delta;
                                stuff.gsa2.bg[1].scroll -= delta;
                                //gsa2.bgs[2].scroll -= delta;
                            }
                            _ => {}
                        }
                    }
                    stuff.mouse_pos = new_pos;
		    let new_tile_pos =
                    IVec2 {
                        x: (new_pos.x + stuff.gsa1.bg[0].scroll.x).div_euclid(TILE_SIZE as i32),
                        y: (new_pos.y + stuff.gsa1.bg[0].scroll.y).div_euclid(TILE_SIZE as i32),
                    };
		    let tile_pos_changed = stuff.tile_pos != new_tile_pos;
		    stuff.tile_pos = new_tile_pos;
                    stuff.tile_pos2 = IVec2 {
                        x: (new_pos.x + stuff.gsa2.bg[0].scroll.x).div_euclid(TILE_SIZE as i32),
                        y: (new_pos.y + stuff.gsa2.bg[0].scroll.y).div_euclid(TILE_SIZE as i32),
                    };
                    let cursor_pos = stuff.tile_pos * TILE_SIZE as i32 - stuff.gsa1.bg[0].scroll;
                    let cursor_pos2 = stuff.tile_pos2 * TILE_SIZE as i32 - stuff.gsa2.bg[0].scroll;
                    stuff.gsa1.sprite[SPR_CURSOR].pos = cursor_pos;
                    stuff.gsa2.sprite[SPR_CURSOR].pos = cursor_pos2;
		    if tile_pos_changed {
			stuff.update_status_bar();
		    }
                }
                _ => {}
            },

            Event::MainEventsCleared => {
                match stuff.state {
                    State::Edit => {
                        if stuff.left_down {
                            if stuff.tile_pos.x >= 0
                                && stuff.tile_pos.y >= 0
                                && stuff.tile_pos.x < BACKGROUND_MAX_SIZE as i32
                                && stuff.tile_pos.y < BACKGROUND_MAX_SIZE as i32
                            {
                                let tile =
                                    (stuff.selected_tile.x + (stuff.selected_tile.y << 8)) as u16;
                                stuff.gsa1.bg[stuff.current_layer].tiles
                                    [stuff.tile_pos.x as usize]
                                    [stuff.tile_pos.y as usize] = tile;
                            }
                        } else if stuff.right_down {
                            if stuff.tile_pos.x >= 0
                                && stuff.tile_pos.y >= 0
                                && stuff.tile_pos.x < BACKGROUND_MAX_SIZE as i32
                                && stuff.tile_pos.y < BACKGROUND_MAX_SIZE as i32
                            {
                                stuff.gsa1.bg[stuff.current_layer].tiles
                                    [stuff.tile_pos.x as usize]
                                    [stuff.tile_pos.y as usize] = EMPTY_TILE;
                            }
                        }
                    }
                    State::SelectTile => {
                        if stuff.left_down {
                            if stuff.tile_pos2.x >= 0
                                && stuff.tile_pos2.y >= 0
                                && stuff.tile_pos2.x < BACKGROUND_MAX_SIZE as i32
                                && stuff.tile_pos2.y < BACKGROUND_MAX_SIZE as i32
                            {
                                stuff.gsa2.bg[1].tiles[stuff.selected_tile.x as usize]
                                    [stuff.selected_tile.y as usize] = EMPTY_TILE;
                                stuff.selected_tile = stuff.tile_pos2;
                                stuff.gsa2.bg[1].tiles[stuff.selected_tile.x as usize]
                                    [stuff.selected_tile.y as usize] = TILE_MARKER;
                            }
                        }
                    }
                    _ => {}
                };

                // render
                let size = window.inner_size();
                surface
                    .resize(
                        NonZeroU32::new(size.width).unwrap(),
                        NonZeroU32::new(size.height).unwrap(),
                    )
                    .unwrap();
                let mut screen_buffer =
                    vec![TRANSPARENT; size.width as usize * size.height as usize / 4];
                let mut window_buffer = surface.buffer_mut().unwrap();
                let screen_size = IVec2 {
                    x: size.width as i32 / 2,
                    y: size.height as i32 / 2,
                };
                render_to_screen(&mut screen_buffer, &stuff.gsa1, &tileset, screen_size);
                render_to_screen(&mut screen_buffer, &stuff.gsa2, &tileset, screen_size);
                let mut screen = Surface {
                    data: &mut screen_buffer,
                    size: screen_size,
                };
                match stuff.state {
                    State::Edit => {
                        let xs = -stuff.gsa1.bg[0].scroll.x - 1;
                        let ys = -stuff.gsa1.bg[0].scroll.y - 1;
                        let xe = xs + stuff.gsa1.bg[0].size.x * TILE_SIZE as i32 + 1;
                        let ye = ys + stuff.gsa1.bg[0].size.y * TILE_SIZE as i32 + 1;
                        if xs >= 16 && xs < screen_size.x {
                            let starty = (ys + 1).max(0);
                            let len = ye.min(screen_size.y - 10) - starty;
                            screen.draw_vline(IVec2 { x: xs, y: starty }, len, 0xfc);
                        }
                        if xe >= 16 && xe < screen_size.x {
                            let starty = (ys + 1).max(0);
                            let len = ye.min(screen_size.y - 10) - starty;
                            screen.draw_vline(IVec2 { x: xe, y: starty }, len, 0xfc);
                        }
                        if ys >= 0 && ys < (screen_size.y - 10) {
                            let startx = (xs + 1).max(16);
                            let len = xe.min(screen_size.x) - startx;
                            screen.draw_hline(IVec2 { x: startx, y: ys }, len, 0xfc);
                        }
                        if ye >= 0 && ye < (screen_size.y - 10) {
                            let startx = (xs + 1).max(16);
                            let len = xe.min(screen_size.x) - startx;
                            screen.draw_hline(IVec2 { x: startx, y: ye }, len, 0xfc);
                        }
                    }
                    _ => {}
                }

                render_to_window(
                    &mut window_buffer,
                    &mut screen_buffer,
                    &palette,
                    IVec2 {
                        x: size.width as i32,
                        y: size.height as i32,
                    },
                    IVec2 {
                        x: size.width as i32 / 2,
                        y: size.height as i32 / 2,
                    },
                    2,
                    0,
                    0,
                );
                window_buffer.present().unwrap();
            }
            _ => {}
        }
    });
}
